package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.lookup.RawTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.ProblemMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBindingVisitor;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.CaptureBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.BinaryTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.Substitution;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;

public class NullAnnotationMatching
{
    public static final NullAnnotationMatching NULL_ANNOTATIONS_OK;
    public static final NullAnnotationMatching NULL_ANNOTATIONS_OK_NONNULL;
    public static final NullAnnotationMatching NULL_ANNOTATIONS_UNCHECKED;
    public static final NullAnnotationMatching NULL_ANNOTATIONS_MISMATCH;
    private final NullAnnotationMatching.Severity severity;
    public final TypeBinding superTypeHint;
    public final int nullStatus;
    private static /* synthetic */ int[] $SWITCH_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode;
    
    static {
        NULL_ANNOTATIONS_OK = new NullAnnotationMatching(NullAnnotationMatching.Severity.OK, 1, null);
        NULL_ANNOTATIONS_OK_NONNULL = new NullAnnotationMatching(NullAnnotationMatching.Severity.OK, 4, null);
        NULL_ANNOTATIONS_UNCHECKED = new NullAnnotationMatching(NullAnnotationMatching.Severity.UNCHECKED, 1, null);
        NULL_ANNOTATIONS_MISMATCH = new NullAnnotationMatching(NullAnnotationMatching.Severity.MISMATCH, 1, null);
    }
    
    NullAnnotationMatching(NullAnnotationMatching.Severity severity, int nullStatus, TypeBinding superTypeHint) {
        this.severity = severity;
        this.superTypeHint = superTypeHint;
        this.nullStatus = nullStatus;
    }
    
    public boolean isAnyMismatch() {
        return this.severity.isAnyMismatch();
    }
    
    public boolean isUnchecked() {
        return this.severity == NullAnnotationMatching.Severity.UNCHECKED;
    }
    
    public boolean isDefiniteMismatch() {
        return this.severity == NullAnnotationMatching.Severity.MISMATCH;
    }
    
    public boolean wantToReport() {
        return this.severity == NullAnnotationMatching.Severity.LEGACY_WARNING;
    }
    
    public boolean isPotentiallyNullMismatch() {
        return !this.isDefiniteMismatch() && this.nullStatus != -1 && (this.nullStatus & 0x10) != 0x0;
    }
    
    public String superTypeHintName(CompilerOptions options, boolean shortNames) {
        return String.valueOf(this.superTypeHint.nullAnnotatedReadableName(options, shortNames));
    }
    
    public static int checkAssignment(BlockScope currentScope, FlowContext flowContext, VariableBinding var, FlowInfo flowInfo, int nullStatus, Expression expression, TypeBinding providedType) {
        if (providedType == null) {
            return 1;
        }
        long lhsTagBits = 0L;
        boolean hasReported = false;
        boolean usesNullTypeAnnotations = currentScope.environment().usesNullTypeAnnotations();
        if (!usesNullTypeAnnotations) {
            lhsTagBits = (var.tagBits & 0x180000000000000L);
        }
        else if (expression instanceof ConditionalExpression && expression.isPolyExpression()) {
            ConditionalExpression ce = (ConditionalExpression)expression;
            int status1 = checkAssignment(currentScope, flowContext, var, flowInfo, ce.ifTrueNullStatus, ce.valueIfTrue, ce.valueIfTrue.resolvedType);
            int status2 = checkAssignment(currentScope, flowContext, var, flowInfo, ce.ifFalseNullStatus, ce.valueIfFalse, ce.valueIfFalse.resolvedType);
            if (status1 == status2) {
                return status1;
            }
            return nullStatus;
        }
        else {
            lhsTagBits = (var.type.tagBits & 0x180000000000000L);
            NullAnnotationMatching annotationStatus = analyse(var.type, providedType, null, null, nullStatus, expression, NullAnnotationMatching.CheckMode.COMPATIBLE);
            if (annotationStatus.isAnyMismatch()) {
                flowContext.recordNullityMismatch(currentScope, expression, providedType, var.type, flowInfo, nullStatus, annotationStatus);
                hasReported = true;
            }
            else {
                if (annotationStatus.wantToReport()) {
                    annotationStatus.report(currentScope);
                }
                if (annotationStatus.nullStatus != 1) {
                    return annotationStatus.nullStatus;
                }
            }
        }
        if (lhsTagBits == 72057594037927936L && nullStatus != 4) {
            if (!hasReported) {
                flowContext.recordNullityMismatch(currentScope, expression, providedType, var.type, flowInfo, nullStatus, null);
            }
            return 4;
        }
        else if (lhsTagBits == 36028797018963968L && nullStatus == 1) {
            if (usesNullTypeAnnotations && providedType.isTypeVariable() && (providedType.tagBits & 0x180000000000000L) == 0x0L) {
                return 48;
            }
            return 24;
        }
        else {
            return nullStatus;
        }
    }
    
    public static NullAnnotationMatching analyse(TypeBinding requiredType, TypeBinding providedType, int nullStatus) {
        return analyse(requiredType, providedType, null, null, nullStatus, null, NullAnnotationMatching.CheckMode.COMPATIBLE);
    }
    
    public static NullAnnotationMatching analyse(TypeBinding requiredType, TypeBinding providedType, TypeBinding providedSubstitute, Substitution substitution, int nullStatus, Expression providedExpression, NullAnnotationMatching.CheckMode mode) {
        if (!requiredType.enterRecursiveFunction()) {
            return NullAnnotationMatching.NULL_ANNOTATIONS_OK;
        }
        try {
            NullAnnotationMatching.Severity severity = NullAnnotationMatching.Severity.OK;
            TypeBinding superTypeHint = null;
            NullAnnotationMatching okStatus = NullAnnotationMatching.NULL_ANNOTATIONS_OK;
            if (areSameTypes(requiredType, providedType, providedSubstitute)) {
                if ((requiredType.tagBits & 0x100000000000000L) != 0x0L) {
                    return okNonNullStatus(providedExpression);
                }
                return okStatus;
            }
            else {
                if (requiredType instanceof TypeVariableBinding && substitution != null && (mode == NullAnnotationMatching.CheckMode.EXACT || mode == NullAnnotationMatching.CheckMode.COMPATIBLE || mode == NullAnnotationMatching.CheckMode.BOUND_SUPER_CHECK)) {
                    requiredType.exitRecursiveFunction();
                    requiredType = Scope.substitute(substitution, requiredType);
                    if (!requiredType.enterRecursiveFunction()) {
                        return NullAnnotationMatching.NULL_ANNOTATIONS_OK;
                    }
                    if (areSameTypes(requiredType, providedType, providedSubstitute)) {
                        if ((requiredType.tagBits & 0x100000000000000L) != 0x0L) {
                            return okNonNullStatus(providedExpression);
                        }
                        return okStatus;
                    }
                }
                if (mode == NullAnnotationMatching.CheckMode.BOUND_CHECK && requiredType instanceof TypeVariableBinding) {
                    boolean passedBoundCheck = substitution instanceof ParameterizedTypeBinding && (((ParameterizedTypeBinding)substitution).tagBits & 0x400000L) != 0x0L;
                    if (!passedBoundCheck) {
                        TypeBinding superClass = requiredType.superclass();
                        if (superClass != null && (superClass.hasNullTypeAnnotations() || substitution != null)) {
                            NullAnnotationMatching status = analyse(superClass, providedType, null, substitution, nullStatus, providedExpression, NullAnnotationMatching.CheckMode.BOUND_SUPER_CHECK);
                            severity = severity.max(status.severity);
                            if (severity == NullAnnotationMatching.Severity.MISMATCH) {
                                return new NullAnnotationMatching(severity, nullStatus, superTypeHint);
                            }
                        }
                        TypeBinding[] superInterfaces = requiredType.superInterfaces();
                        if (superInterfaces != null) {
                            for (int i = 0; i < superInterfaces.length; ++i) {
                                if (superInterfaces[i].hasNullTypeAnnotations() || substitution != null) {
                                    NullAnnotationMatching status2 = analyse(superInterfaces[i], providedType, null, substitution, nullStatus, providedExpression, NullAnnotationMatching.CheckMode.BOUND_SUPER_CHECK);
                                    severity = severity.max(status2.severity);
                                    if (severity == NullAnnotationMatching.Severity.MISMATCH) {
                                        return new NullAnnotationMatching(severity, nullStatus, superTypeHint);
                                    }
                                }
                            }
                        }
                    }
                }
                if ((requiredType instanceof ArrayBinding)) {
                    long[] requiredDimsTagBits = ((ArrayBinding)requiredType).nullTagBitsPerDimension;
                    if (requiredDimsTagBits != null) {
                        int dims = requiredType.dimensions();
                        if (requiredType.dimensions() == providedType.dimensions()) {
                            long[] providedDimsTagBits = ((ArrayBinding)providedType).nullTagBitsPerDimension;
                            if (providedDimsTagBits == null) {
                                providedDimsTagBits = new long[dims + 1];
                            }
                            int currentNullStatus = nullStatus;
                            for (int k = 0; k <= dims; ++k) {
                                long requiredBits2 = validNullTagBits(requiredDimsTagBits[k]);
                                long providedBits2 = validNullTagBits(providedDimsTagBits[k]);
                                if (k == 0 && requiredBits2 == 36028797018963968L && nullStatus != -1 && mode.requiredNullableMatchesAll()) {
                                    if (nullStatus == 2) {
                                        break;
                                    }
                                }
                                else {
                                    if (k > 0) {
                                        currentNullStatus = -1;
                                    }
                                    NullAnnotationMatching.Severity dimSeverity = computeNullProblemSeverity(requiredBits2, providedBits2, currentNullStatus, (k == 0) ? mode : mode.toDetail(), false);
                                    if (k > 0 && dimSeverity == NullAnnotationMatching.Severity.UNCHECKED && providedExpression instanceof ArrayAllocationExpression && providedBits2 == 0L && requiredBits2 != 0L) {
                                        Expression[] dimensions = ((ArrayAllocationExpression)providedExpression).dimensions;
                                        Expression previousDim = dimensions[k - 1];
                                        if (previousDim instanceof IntLiteral && previousDim.constant.intValue() == 0) {
                                            dimSeverity = NullAnnotationMatching.Severity.OK;
                                            nullStatus = -1;
                                            break;
                                        }
                                    }
                                    severity = severity.max(dimSeverity);
                                    if (severity == NullAnnotationMatching.Severity.MISMATCH) {
                                        return NullAnnotationMatching.NULL_ANNOTATIONS_MISMATCH;
                                    }
                                }
                                if (severity == NullAnnotationMatching.Severity.OK) {
                                    nullStatus = -1;
                                }
                            }
                        }
                        else if (providedType.id == 12 && dims > 0 && requiredDimsTagBits[0] == 72057594037927936L) {
                            return NullAnnotationMatching.NULL_ANNOTATIONS_MISMATCH;
                        }
                    }
                }
                else if (requiredType.hasNullTypeAnnotations() || providedType.hasNullTypeAnnotations() || requiredType.isTypeVariable()) {
                    long requiredBits = requiredNullTagBits(requiredType, mode);
                    if (requiredBits != 36028797018963968L || nullStatus == -1 || !mode.requiredNullableMatchesAll()) {
                        long providedBits = providedNullTagBits(providedType);
                        NullAnnotationMatching.Severity s = computeNullProblemSeverity(requiredBits, providedBits, nullStatus, mode, requiredType.isTypeVariable());
                        if (s.isAnyMismatch() && requiredType.isWildcard() && requiredBits != 0L && ((WildcardBinding)requiredType).determineNullBitsFromDeclaration(null, null) == 0L) {
                            s = NullAnnotationMatching.Severity.OK;
                        }
                        severity = severity.max(s);
                        if (!severity.isAnyMismatch() && (providedBits & 0x100000000000000L) != 0x0L) {
                            okStatus = okNonNullStatus(providedExpression);
                        }
                    }
                    if (severity != NullAnnotationMatching.Severity.MISMATCH && nullStatus != 2) {
                        TypeBinding providedSuper = providedType.findSuperTypeOriginatingFrom(requiredType);
                        TypeBinding providedSubstituteSuper = (providedSubstitute != null) ? providedSubstitute.findSuperTypeOriginatingFrom(requiredType) : null;
                        if (severity == NullAnnotationMatching.Severity.UNCHECKED && requiredType.isTypeVariable() && providedType.isTypeVariable() && (providedSuper == requiredType || providedSubstituteSuper == requiredType)) {
                            severity = NullAnnotationMatching.Severity.OK;
                        }
                        if (providedSuper != providedType) {
                            superTypeHint = providedSuper;
                        }
                        if (requiredType.isParameterizedType() && providedSuper instanceof ParameterizedTypeBinding) {
                            TypeBinding[] requiredArguments = ((ParameterizedTypeBinding)requiredType).arguments;
                            TypeBinding[] providedArguments = ((ParameterizedTypeBinding)providedSuper).arguments;
                            TypeBinding[] providedSubstitutes = (providedSubstituteSuper instanceof ParameterizedTypeBinding) ? ((ParameterizedTypeBinding)providedSubstituteSuper).arguments : null;
                            if (requiredArguments != null && providedArguments != null && requiredArguments.length == providedArguments.length) {
                                int j = 0;
                                while (j < requiredArguments.length) {
                                    TypeBinding providedArgSubstitute = (providedSubstitutes != null) ? providedSubstitutes[j] : null;
                                    NullAnnotationMatching status3 = analyse(requiredArguments[j], providedArguments[j], providedArgSubstitute, substitution, -1, providedExpression, mode.toDetail());
                                    severity = severity.max(status3.severity);
                                    if (severity == NullAnnotationMatching.Severity.MISMATCH) {
                                        return new NullAnnotationMatching(severity, nullStatus, superTypeHint);
                                    }
                                    ++j;
                                }
                            }
                        }
                        TypeBinding requiredEnclosing = requiredType.enclosingType();
                        TypeBinding providedEnclosing = providedType.enclosingType();
                        if (requiredEnclosing != null && providedEnclosing != null) {
                            TypeBinding providedEnclSubstitute = (providedSubstitute != null) ? providedSubstitute.enclosingType() : null;
                            NullAnnotationMatching status4 = analyse(requiredEnclosing, providedEnclosing, providedEnclSubstitute, substitution, -1, providedExpression, mode);
                            severity = severity.max(status4.severity);
                        }
                    }
                }
                if (!severity.isAnyMismatch()) {
                    return okStatus;
                }
                return new NullAnnotationMatching(severity, nullStatus, superTypeHint);
            }
        }
        finally {
            requiredType.exitRecursiveFunction();
        }
    }
    
    public void report(Scope scope) {
    }
    
    public static NullAnnotationMatching okNonNullStatus(final Expression providedExpression) {
        if ((providedExpression instanceof MessageSend)) {
            final MethodBinding method = ((MessageSend)providedExpression).binding;
            if (method != null && method.isValidBinding()) {
                MethodBinding originalMethod = method.original();
                TypeBinding originalDeclaringClass = originalMethod.declaringClass;
                if (originalDeclaringClass instanceof BinaryTypeBinding && ((BinaryTypeBinding)originalDeclaringClass).externalAnnotationStatus.isPotentiallyUnannotatedLib() && originalMethod.returnType.isTypeVariable() && (originalMethod.returnType.tagBits & 0x180000000000000L) == 0x0L) {
                    final int severity = (((BinaryTypeBinding)originalDeclaringClass).externalAnnotationStatus == BinaryTypeBinding.ExternalAnnotationStatus.NO_EEA_FILE) ? 0 : 1024;
                    return new NullAnnotationMatching(NullAnnotationMatching.Severity.LEGACY_WARNING, 1, null) {
                        @Override
                        public void report(Scope scope) {
                            scope.problemReporter().nonNullTypeVariableInUnannotatedBinary(scope.environment(), method, providedExpression, severity);
                        }
                    };
                }
            }
        }
        return NullAnnotationMatching.NULL_ANNOTATIONS_OK_NONNULL;
    }
    
    protected static boolean areSameTypes(TypeBinding requiredType, TypeBinding providedType, TypeBinding providedSubstitute) {
        if (requiredType == providedType) {
            return true;
        }
        if (requiredType.isParameterizedType() || requiredType.isArrayType()) {
            return false;
        }
        if (TypeBinding.notEquals(requiredType, providedType)) {
            if ((requiredType instanceof CaptureBinding)) {
                TypeBinding lowerBound = ((CaptureBinding)requiredType).lowerBound;
                if (lowerBound != null && areSameTypes(lowerBound, providedType, providedSubstitute)) {
                    return (requiredType.tagBits & 0x180000000000000L) == (providedType.tagBits & 0x180000000000000L);
                }
            }
            else {
                if (requiredType.kind() == 4100 && requiredType == providedSubstitute) {
                    return true;
                }
                if ((providedType instanceof CaptureBinding)) {
                    TypeBinding upperBound = ((CaptureBinding)providedType).upperBound();
                    if (upperBound != null && areSameTypes(requiredType, upperBound, providedSubstitute)) {
                        return (requiredType.tagBits & 0x180000000000000L) == (providedType.tagBits & 0x180000000000000L);
                    }
                }
            }
            return false;
        }
        else {
            return (requiredType.tagBits & 0x180000000000000L) == (providedType.tagBits & 0x180000000000000L);
        }
    }
    
    static long requiredNullTagBits(TypeBinding type, NullAnnotationMatching.CheckMode mode) {
        long tagBits = type.tagBits & 0x180000000000000L;
        if (tagBits != 0L) {
            return validNullTagBits(tagBits);
        }
        if (type.isWildcard()) {
            return 108086391056891904L;
        }
        if (type.isTypeVariable()) {
            if (type.isCapture()) {
                TypeBinding lowerBound = ((CaptureBinding)type).lowerBound;
                if (lowerBound != null) {
                    tagBits = (lowerBound.tagBits & 0x180000000000000L);
                    if (tagBits == 36028797018963968L) {
                        return 36028797018963968L;
                    }
                }
            }
            switch ($SWITCH_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode()[mode.ordinal()]) {
                case 3:
                case 4:
                case 5:
                case 6: {
                    break;
                }
                default: {
                    return 72057594037927936L;
                }
            }
        }
        return 0L;
    }
    
    static long providedNullTagBits(TypeBinding type) {
        long tagBits = type.tagBits & 0x180000000000000L;
        if (tagBits != 0L) {
            return validNullTagBits(tagBits);
        }
        if (type.isWildcard()) {
            return 108086391056891904L;
        }
        if (type.isTypeVariable()) {
            TypeVariableBinding typeVariable = (TypeVariableBinding)type;
            boolean haveNullBits = false;
            if (typeVariable.isCapture()) {
                TypeBinding lowerBound = ((CaptureBinding)typeVariable).lowerBound;
                if (lowerBound != null) {
                    tagBits = (lowerBound.tagBits & 0x180000000000000L);
                    if (tagBits == 36028797018963968L) {
                        return 36028797018963968L;
                    }
                    haveNullBits |= (tagBits != 0L);
                }
            }
            if (typeVariable.firstBound != null) {
                long boundBits = typeVariable.firstBound.tagBits & 0x180000000000000L;
                if (boundBits == 72057594037927936L) {
                    return 72057594037927936L;
                }
                haveNullBits |= (boundBits != 0L);
            }
            if (haveNullBits) {
                return 108086391056891904L;
            }
        }
        return 0L;
    }
    
    public static long validNullTagBits(long bits) {
        bits &= 0x180000000000000L;
        return (bits == 108086391056891904L) ? 0L : bits;
    }
    
    public static TypeBinding moreDangerousType(TypeBinding one, TypeBinding two) {
        if (one == null) {
            return null;
        }
        long oneNullBits = validNullTagBits(one.tagBits);
        long twoNullBits = validNullTagBits(two.tagBits);
        if (oneNullBits != twoNullBits) {
            if (oneNullBits == 36028797018963968L) {
                return one;
            }
            if (twoNullBits == 36028797018963968L) {
                return two;
            }
            if (oneNullBits == 0L) {
                return one;
            }
            return two;
        }
        else {
            if (one != two && analyse(one, two, -1).isAnyMismatch()) {
                return two;
            }
            return one;
        }
    }
    
    private static NullAnnotationMatching.Severity computeNullProblemSeverity(long requiredBits, long providedBits, int nullStatus, NullAnnotationMatching.CheckMode mode, boolean requiredIsTypeVariable) {
        if (requiredBits == providedBits) {
            return NullAnnotationMatching.Severity.OK;
        }
        if (requiredBits == 0L) {
            switch ($SWITCH_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode()[mode.ordinal()]) {
                case 1:
                case 2:
                case 3:
                case 4: {
                    return NullAnnotationMatching.Severity.OK;
                }
                case 5: {
                    if (providedBits == 72057594037927936L) {
                        return NullAnnotationMatching.Severity.OK;
                    }
                    if (!requiredIsTypeVariable) {
                        return NullAnnotationMatching.Severity.OK;
                    }
                    return NullAnnotationMatching.Severity.UNCHECKED;
                }
                case 6: {
                    return NullAnnotationMatching.Severity.UNCHECKED;
                }
                default: {
                    break;
                }
            }
        }
        else {
            if (requiredBits == 108086391056891904L) {
                return NullAnnotationMatching.Severity.OK;
            }
            if (requiredBits == 72057594037927936L) {
                switch ($SWITCH_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode()[mode.ordinal()]) {
                    case 1:
                    case 4: {
                        if (nullStatus == 4) {
                            return NullAnnotationMatching.Severity.OK;
                        }
                    }
                    case 2:
                    case 3:
                    case 5:
                    case 6: {
                        if (providedBits == 0L) {
                            return NullAnnotationMatching.Severity.UNCHECKED;
                        }
                        return NullAnnotationMatching.Severity.MISMATCH;
                    }
                    default: {
                        break;
                    }
                }
            }
            else if (requiredBits == 36028797018963968L) {
                switch ($SWITCH_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode()[mode.ordinal()]) {
                    case 1:
                    case 4:
                    case 5: {
                        return NullAnnotationMatching.Severity.OK;
                    }
                    case 2:
                    case 3: {
                        if (providedBits == 0L) {
                            return NullAnnotationMatching.Severity.UNCHECKED;
                        }
                        return NullAnnotationMatching.Severity.MISMATCH;
                    }
                    case 6: {
                        return NullAnnotationMatching.Severity.MISMATCH;
                    }
                    default: {
                        break;
                    }
                }
            }
        }
        return NullAnnotationMatching.Severity.OK;
    }
    
    public static MethodBinding checkForContradictions(MethodBinding method, Object location, Scope scope) {
        int start = 0;
        int end = 0;
        if ((location instanceof InvocationSite)) {
            start = ((InvocationSite)location).sourceStart();
            end = ((InvocationSite)location).sourceEnd();
        }
        else if ((location instanceof ASTNode)) {
            start = ((ASTNode)location).sourceStart;
            end = ((ASTNode)location).sourceEnd;
        }
        NullAnnotationMatching.SearchContradictions searchContradiction = new NullAnnotationMatching.SearchContradictions();
        TypeBindingVisitor.visit(searchContradiction, method.returnType);
        if (searchContradiction.typeWithContradiction != null) {
            if (scope == null) {
                return new ProblemMethodBinding(method, method.selector, method.parameters, 25);
            }
            scope.problemReporter().contradictoryNullAnnotationsInferred(method, start, end, location instanceof FunctionalExpression);
            return method;
        }
        else {
            Expression[] arguments = null;
            if ((location instanceof Invocation)) {
                arguments = ((Invocation)location).arguments();
            }
            int i = 0;
            while (i < method.parameters.length) {
                TypeBindingVisitor.visit(searchContradiction, method.parameters[i]);
                if (searchContradiction.typeWithContradiction != null) {
                    if (scope == null) {
                        return new ProblemMethodBinding(method, method.selector, method.parameters, 25);
                    }
                    if (arguments != null && i < arguments.length) {
                        scope.problemReporter().contradictoryNullAnnotationsInferred(method, arguments[i]);
                        return method;
                    }
                    scope.problemReporter().contradictoryNullAnnotationsInferred(method, start, end, location instanceof FunctionalExpression);
                    return method;
                }
                else {
                    ++i;
                }
            }
            return method;
        }
    }
    
    public static boolean hasContradictions(TypeBinding type) {
        NullAnnotationMatching.SearchContradictions searchContradiction = new NullAnnotationMatching.SearchContradictions();
        TypeBindingVisitor.visit(searchContradiction, type);
        return searchContradiction.typeWithContradiction != null;
    }
    
    public static TypeBinding strongerType(TypeBinding type1, TypeBinding type2, LookupEnvironment environment) {
        if ((type1.tagBits & 0x100000000000000L) != 0x0L) {
            return mergeTypeAnnotations(type1, type2, true, environment);
        }
        return mergeTypeAnnotations(type2, type1, true, environment);
    }
    
    public static TypeBinding[] weakerTypes(TypeBinding[] parameters1, TypeBinding[] parameters2, LookupEnvironment environment) {
        TypeBinding[] newParameters = new TypeBinding[parameters1.length];
        for (int i = 0; i < newParameters.length; ++i) {
            long tagBits1 = parameters1[i].tagBits;
            long tagBits2 = parameters2[i].tagBits;
            if ((tagBits1 & 0x80000000000000L) != 0x0L) {
                newParameters[i] = mergeTypeAnnotations(parameters1[i], parameters2[i], true, environment);
            }
            else if ((tagBits2 & 0x80000000000000L) != 0x0L) {
                newParameters[i] = mergeTypeAnnotations(parameters2[i], parameters1[i], true, environment);
            }
            else if ((tagBits1 & 0x100000000000000L) == 0x0L) {
                newParameters[i] = mergeTypeAnnotations(parameters1[i], parameters2[i], true, environment);
            }
            else {
                newParameters[i] = mergeTypeAnnotations(parameters2[i], parameters1[i], true, environment);
            }
        }
        return newParameters;
    }
    
    private static TypeBinding mergeTypeAnnotations(TypeBinding type, TypeBinding otherType, boolean top, LookupEnvironment environment) {
        TypeBinding mainType = type;
        if (!top) {
            AnnotationBinding[] otherAnnotations = otherType.getTypeAnnotations();
            if (otherAnnotations != Binding.NO_ANNOTATIONS) {
                mainType = environment.createAnnotatedType(type, otherAnnotations);
            }
        }
        if (mainType.isParameterizedType() && otherType.isParameterizedType()) {
            ParameterizedTypeBinding ptb = (ParameterizedTypeBinding)type;
            ParameterizedTypeBinding otherPTB = (ParameterizedTypeBinding)otherType;
            TypeBinding[] typeArguments = ptb.arguments;
            TypeBinding[] otherTypeArguments = otherPTB.arguments;
            TypeBinding[] newTypeArguments = new TypeBinding[typeArguments.length];
            for (int i = 0; i < typeArguments.length; ++i) {
                newTypeArguments[i] = mergeTypeAnnotations(typeArguments[i], otherTypeArguments[i], false, environment);
            }
            return environment.createParameterizedType(ptb.genericType(), newTypeArguments, ptb.enclosingType());
        }
        return mainType;
    }
    
    @Override
    public String toString() {
        if (this == NullAnnotationMatching.NULL_ANNOTATIONS_OK) {
            return "OK";
        }
        if (this == NullAnnotationMatching.NULL_ANNOTATIONS_MISMATCH) {
            return "MISMATCH";
        }
        if (this == NullAnnotationMatching.NULL_ANNOTATIONS_OK_NONNULL) {
            return "OK NonNull";
        }
        if (this == NullAnnotationMatching.NULL_ANNOTATIONS_UNCHECKED) {
            return "UNCHECKED";
        }
        StringBuilder buf = new StringBuilder();
        buf.append("Analysis result: severity=" + this.severity);
        buf.append(" nullStatus=" + this.nullStatus);
        return buf.toString();
    }
    
    static /* synthetic */ int[] $SWITCH_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode() {
        int[] $switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode = NullAnnotationMatching.$SWITCH_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode;
        if ($switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode != null) {
            return $switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode;
        }
        int[] $switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode2 = new int[NullAnnotationMatching.CheckMode.values().length];
        try {
            $switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode2[NullAnnotationMatching.CheckMode.BOUND_CHECK.ordinal()] = 3;
        }
        catch (NoSuchFieldError noSuchFieldError) {}
        try {
            $switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode2[NullAnnotationMatching.CheckMode.BOUND_SUPER_CHECK.ordinal()] = 4;
        }
        catch (NoSuchFieldError noSuchFieldError2) {}
        try {
            $switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode2[NullAnnotationMatching.CheckMode.COMPATIBLE.ordinal()] = 1;
        }
        catch (NoSuchFieldError noSuchFieldError3) {}
        try {
            $switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode2[NullAnnotationMatching.CheckMode.EXACT.ordinal()] = 2;
        }
        catch (NoSuchFieldError noSuchFieldError4) {}
        try {
            $switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode2[NullAnnotationMatching.CheckMode.OVERRIDE.ordinal()] = 6;
        }
        catch (NoSuchFieldError noSuchFieldError5) {}
        try {
            $switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode2[NullAnnotationMatching.CheckMode.OVERRIDE_RETURN.ordinal()] = 5;
        }
        catch (NoSuchFieldError noSuchFieldError6) {}
        return NullAnnotationMatching.$SWITCH_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode = $switch_TABLE$org$eclipse$jdt$internal$compiler$ast$NullAnnotationMatching$CheckMode2;
    }
    
    public enum CheckMode
    {
        COMPATIBLE {
            @Override
            boolean requiredNullableMatchesAll() {
                return true;
            }
        }, 
        EXACT("EXACT", 1), 
        BOUND_CHECK("BOUND_CHECK", 2), 
        BOUND_SUPER_CHECK("BOUND_SUPER_CHECK", 3), 
        OVERRIDE_RETURN {
            @Override
            NullAnnotationMatching.CheckMode toDetail() {
                return NullAnnotationMatching.CheckMode.2.OVERRIDE;
            }
        }, 
        OVERRIDE {
            @Override
            boolean requiredNullableMatchesAll() {
                return true;
            }
            
            @Override
            NullAnnotationMatching.CheckMode toDetail() {
                return NullAnnotationMatching.CheckMode.3.OVERRIDE;
            }
        };
        
        private CheckMode(String name, int ordinal) {
        }
        
        boolean requiredNullableMatchesAll() {
            return false;
        }
        
        NullAnnotationMatching.CheckMode toDetail() {
            return NullAnnotationMatching.CheckMode.EXACT;
        }
    }
    
    private enum Severity
    {
        OK("OK", 0), 
        LEGACY_WARNING("LEGACY_WARNING", 1), 
        UNCHECKED("UNCHECKED", 2), 
        MISMATCH("MISMATCH", 3);
        
        private Severity(String name, int ordinal) {
        }
        
        public NullAnnotationMatching.Severity max(NullAnnotationMatching.Severity severity) {
            if (this.compareTo(severity) < 0) {
                return severity;
            }
            return this;
        }
        
        public boolean isAnyMismatch() {
            return this.compareTo(NullAnnotationMatching.Severity.LEGACY_WARNING) > 0;
        }
    }
    
    static class SearchContradictions extends TypeBindingVisitor
    {
        ReferenceBinding typeWithContradiction;
        
        public boolean visit(ReferenceBinding referenceBinding) {
            if ((referenceBinding.tagBits & 0x180000000000000L) == 0x180000000000000L) {
                this.typeWithContradiction = referenceBinding;
                return false;
            }
            return true;
        }
        
        public boolean visit(TypeVariableBinding typeVariable) {
            if (!this.visit(typeVariable)) {
                return false;
            }
            long allNullBits = typeVariable.tagBits & 0x180000000000000L;
            if (typeVariable.firstBound != null) {
                allNullBits = (typeVariable.firstBound.tagBits & 0x180000000000000L);
            }
            TypeBinding[] otherUpperBounds;
            for (int length = (otherUpperBounds = typeVariable.otherUpperBounds()).length, i = 0; i < length; ++i) {
                TypeBinding otherBound = otherUpperBounds[i];
                allNullBits |= (otherBound.tagBits & 0x180000000000000L);
            }
            if (allNullBits == 108086391056891904L) {
                this.typeWithContradiction = typeVariable;
                return false;
            }
            return true;
        }
        
        public boolean visit(RawTypeBinding rawType) {
            return this.visit(rawType);
        }
        
        public boolean visit(WildcardBinding wildcardBinding) {
            long allNullBits = wildcardBinding.tagBits & 0x180000000000000L;
            switch (wildcardBinding.boundKind) {
                case 1: {
                    allNullBits |= (wildcardBinding.bound.tagBits & 0x100000000000000L);
                    break;
                }
                case 2: {
                    allNullBits |= (wildcardBinding.bound.tagBits & 0x80000000000000L);
                    break;
                }
            }
            if (allNullBits == 108086391056891904L) {
                this.typeWithContradiction = wildcardBinding;
                return false;
            }
            return true;
        }
        
        public boolean visit(ParameterizedTypeBinding parameterizedTypeBinding) {
            return this.visit(parameterizedTypeBinding) && super.visit(parameterizedTypeBinding);
        }
    }
}
